#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#define xalloc(_t)	((_t*)malloc(sizeof(_t)))
#define RVAL(_x)	((void)0, (_x))

enum move_result {
  MOVE_UNDEFINED,
  MOVE_ILLEGAL,
  MOVE_WIN,
  MOVE_OK
};

typedef enum move_result move_result;

#define DEPTH		5

#define EMPTY_CHAR	'.'
#define HUMAN_CHAR	'x'
#define COMP_CHAR	'o'

#define NROWS	7
#define NCOLS	7
#define WINNUM	4	/* number in a row to win */
#define OFFSET	(WINNUM-1)

typedef struct board *	board;

struct board {
  char	array[NROWS+2*OFFSET][NCOLS+2*OFFSET];
};

#define deref_board(_b)			(_b)
#define x_raw_pos_board(_b, _i, _j)		(deref_board(_b)->array[_i][_j])
#define raw_pos_board(_b, _i, _j)		RVAL(x_raw_pos_board(_b, _i, _j))
#define x_pos_board(_b, _i, _j)	x_raw_pos_board(_b, (_i)+OFFSET, (_j)+OFFSET)
#define pos_board(_b, _i, _j)	RVAL(x_pos_board(_b, _i, _j))

static board
make_board( void )
{
  board	bd = xalloc(struct board);
  int	i;
  int	j;

  for( i = 0; i < NROWS+2*OFFSET; i++ ) {
    for( j = 0; j < NCOLS+2*OFFSET; j++ ) {
      x_raw_pos_board(bd, i, j) = EMPTY_CHAR;
    }
  }
  return bd;
}
	

static void
print_board(
	FILE *	fp,
	board	bd
	)
{
  int	i;
  int	j;

  fprintf(fp, "-");
  for( i = 0; i < NCOLS; i++ ) {
    fprintf(fp, "%d", i);
  }
  fputc('\n', fp);

  for( i = NROWS-1; i >= 0; i-- ) {
    fprintf(fp, "%d", i);
    for( j = 0; j < NCOLS; j++ ) {
      fprintf(fp, "%c", pos_board(bd, i, j));
    }
    fputc('\n', fp);
  }
  fputc('\n', fp);
}

static int
is_win_board(
	board	bd,
	int	row,
	int	col,
	int	play
	)
{
  int	i;
  int	j;
  int	win;

  /* horizontal */
  for( i = 0; i < WINNUM; i++ ) {
    win = 1;
    for( j = 0; j < WINNUM; j++ ) {
      if( pos_board(bd, row, col-i+j) != play ) {
        win = 0;
	break;
      }
    }
    if( win ) {
      return 1;
    }
  }

  /* vertical */
  for( i = 0; i < WINNUM; i++ ) {
    win = 1;
    for( j = 0; j < WINNUM; j++ ) {
      if( pos_board(bd, row-i+j, col) != play ) {
        win = 0;
	break;
      }
    }
    if( win ) {
      return 1;
    }
  }

  /* diagonal up */
  for( i = 0; i < WINNUM; i++ ) {
    win = 1;
    for( j = 0; j < WINNUM; j++ ) {
      if( pos_board(bd, row-i+j, col-i+j) != play ) {
        win = 0;
	break;
      }
    }
    if( win ) {
      return 1;
    }
  }

  /* diagonal down */
  for( i = 0; i < WINNUM; i++ ) {
    win = 1;
    for( j = 0; j < WINNUM; j++ ) {
      if( pos_board(bd, row-i+j, col+i-j) != play ) {
        win = 0;
	break;
      }
    }
    if( win ) {
      return 1;
    }
  }
  return 0;
}


static move_result
add_board(
	board	bd,
	int	col,
	char	play
	)
{
  int	i;

  if( pos_board(bd, NROWS-1, col) != EMPTY_CHAR ) {
    return MOVE_ILLEGAL;
  }

  for( i = NROWS-1; i > 0; i-- ) {
    if( pos_board(bd, i-1, col) != EMPTY_CHAR ) {
      break;
    }
  }
  x_pos_board(bd, i, col) = play;

  if( is_win_board(bd, i, col, play) ) {
    return MOVE_WIN;
  }
  return MOVE_OK;
}

static int
evaluate_max(
	board	bd,
	char	my_play,
	char	other_play,
	int	beta,
	int	ply,
	int *	p_best
	);

static int
evaluate_min(
	board	bd,
	char	my_play,
	char	other_play,
	int	alpha,
	int	ply,
	int *	p_best
	)
{
  int	beta;
  int	score;
  int	col;
  int	i;
  int	best = -1;

  if( ply == 0 ) {
    /* need to do something more intelligent here! */
    return 0;
  }

  beta = INT_MAX;
  for( col = 0; col < NCOLS; col++ ) {
    if( pos_board(bd, NROWS-1, col) != EMPTY_CHAR ) {
      continue;
    }

    for( i = NROWS-1; i > 0; i-- ) {
      if( pos_board(bd, i-1, col) != EMPTY_CHAR ) {
	break;
      }
    }

    x_pos_board(bd, i, col) = other_play;
    if( is_win_board(bd, i, col, other_play) ) {
      /* you've lost; just return */
      x_pos_board(bd, i, col) = EMPTY_CHAR;
      *p_best = col;
      return INT_MIN+1;
    }
    else {
      score = evaluate_max(bd, my_play, other_play, beta, ply-1, p_best);
      if( score < alpha ) {
	x_pos_board(bd, i, col) = EMPTY_CHAR;
	*p_best = col;
        return score;
      }
      if( score < beta ) {
        beta = score;
	best = col;
      }
    }
    x_pos_board(bd, i, col) = EMPTY_CHAR;
  }
  *p_best = best;
  return beta;
}

static int
evaluate_max(
	board	bd,
	char	my_play,
	char	other_play,
	int	beta,
	int	ply,
	int *	p_best
	)
{
  int	alpha;
  int	score;
  int	col;
  int	i;
  int	best = -1;

  if( ply == 0 ) {
    /* need to do something more intelligent here! */
    return 0;
  }

  alpha = INT_MIN;
  for( col = 0; col < NCOLS; col++ ) {
    if( pos_board(bd, NROWS-1, col) != EMPTY_CHAR ) {
      continue;
    }

    for( i = NROWS-1; i > 0; i-- ) {
      if( pos_board(bd, i-1, col) != EMPTY_CHAR ) {
	break;
      }
    }

    x_pos_board(bd, i, col) = my_play;
    if( is_win_board(bd, i, col, my_play) ) {
      /* you've won; just return */
      x_pos_board(bd, i, col) = EMPTY_CHAR; /* undo the move */
      *p_best = col;
      return INT_MAX-1;
    }
    else {
      score = evaluate_min(bd, my_play, other_play, alpha, ply-1, p_best);
      if( score > beta ) {
	/* cut-off */
	x_pos_board(bd, i, col) = EMPTY_CHAR;
	*p_best = col;
        return score;
      }
      if( score > alpha ) {
        alpha = score;
	best = col;
      }
    }
    x_pos_board(bd, i, col) = EMPTY_CHAR;
  }
  *p_best = best;
  return alpha;
}

static int
get_computer_move(
	board	bd
	)
{
  int	best;
  int	score;
  score = evaluate_max(bd, COMP_CHAR, HUMAN_CHAR, INT_MAX, DEPTH, &best);

  return best;
}

static int
is_drawn_board(
	board	bd
	)
{
  int col;
  for( col = 0; col < NCOLS;col++ ) {
    if( pos_board(bd, NROWS-1, col) == EMPTY_CHAR ) {
      return 0;
    }
  }
  return 1;
}

static void
play_loop(
	board	bd
	)
{
  char		c;
  char		dummy;
  char		col;
  move_result	status;

  while( 1 ) {

get_human_move:
    printf("your move (0..%d): ", NCOLS-1);

    c = fgetc(stdin);
    do {
      dummy = fgetc(stdin);
    } while( dummy != '\n' && dummy != EOF);

    if( c == 'q' || c == EOF ) {
      printf("ok, quitting\n");
      return;
    }
    col = c - '0';
    if( col < 0 || col >= NCOLS ) {
      printf("sorry, must be between 0 and %d; try again\n", NCOLS-1);
      goto get_human_move;
    }

    status = add_board(bd, col, HUMAN_CHAR);
    if( status == MOVE_ILLEGAL ) {
      printf("sorry, bad move; try again\n");
      goto get_human_move;
    }
    print_board(stdout, bd);
    if( status == MOVE_WIN ) {
      printf("you win!\n");
      return;
    }

    if( is_drawn_board(bd) ) {
      printf("sorry, drawn game\n");
      return;
    }

    /* get computer move */
    col = get_computer_move(bd);
    printf("computer: %d\n", col);
    status = add_board(bd, col, COMP_CHAR );

    print_board(stdout, bd);
    if( status == MOVE_WIN ) {
      printf("sorry, computer wins\n");
      return;
    }

    if( is_drawn_board(bd) ) {
      printf("sorry, drawn game\n");
      return;
    }
  }
}

int
main(
	int	argc,
	char *	argv[]
	)
{
  board	bd;
  (void)argc;
  (void)argv;

  bd = make_board();
  play_loop(bd);
  return EXIT_SUCCESS;
}
